/* arch-tag: c85f7f4d-d561-40f3-b525-52b6b64c68f4 */


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdbool.h>

#include "pane.h"
#include "x11.h"
#include "util.h"
#include "defaults.h"

struct pane {
        struct {
                int x1,y1;
                int x2,y2;
        } rr; // region needing redrawing;
        bool show_cursor : 1;
        bool redraw_all : 1;
        int cursor_x, cursor_y;
        int width, height;
        int cwidth, cheight;
        struct cell *cells;
        GC gc;
        unsigned long cursor_color;
};


#define CELL(s,x,y) ((s)->cells[(y)*(s)->width + (x)])
#define TOUCHED(s,x,y) ((s)->touched[(y)*(s)->width + (x)])


static void pane_sane(struct pane *p);


struct pane *
pane_create(void)
{
        struct pane *s = malloc(sizeof(struct pane));
        s->cells = NULL;
        s->width = 0;
        s->height = 0;
        s->cursor_x = 0;
        s->cursor_y = 0;
        s->show_cursor = true;
        s->gc = DEF_GC;
        s->cursor_color = DC_cursorColor;
        get_charcell_size(&s->cwidth, &s->cheight);
        s->rr.x1 = s->rr.x2 = -1;
        s->rr.y1 = s->rr.y2 = -1;
        return s;
}

void
pane_setup(struct pane *p, int w, int h) 
{
        p->cells = realloc(p->cells, w*h*sizeof(struct cell));
        memset(p->cells, 0, w * h * sizeof(struct cell));
        p->width = w;
        p->height = h;
        p->cursor_x = 0;
        p->cursor_y = 0;
        pane_touchcell(p,0,0);
        pane_touchcell(p,p->width,p->height);
}

void
pane_settext_latin1(struct pane * restrict p, int x, int y, char const * restrict s, int len, struct cell style)
{
        int i;
        assert(x >= 0 && y >= 0);
        if(len <= 0) return;
        pane_touchcell(p, x, y);
        pane_touchcell(p, MIN(x + len, p->width), y);
        if (y >= p->height) return;
        for(i = 0; i < len && x + i <= p->width; i++) {
                CELL(p, x + i, y) = style;
                CELL(p, x + i, y).ch = s[i];
        }
}

void
pane_settext(struct pane * restrict p, int x, int y, wchar_t const * restrict s, int len, struct cell style)
{
        int i;
        assert(x >= 0 && y >= 0);
        if(len <= 0) return;
        pane_touchcell(p, x, y);
        pane_touchcell(p, MIN(x + len, p->width), y);
        if (y >= p->height) return;
        for(i = 0; i < len && x + i <= p->width; i++) {
                CELL(p, x + i, y) = style;
                CELL(p, x + i, y).ch = s[i];
        }
}

void 
pane_clear(struct pane *p, int x, int y, int w, int h, struct cell style)
{
        int i;
        assert(w >= 0 && h >= 0 && x >= 0 && y >= 0);
        if( w == 0 || h == 0) return;
        pane_touchcell(p, x, y);
        pane_touchcell(p, MIN(x + w, p->width), MIN(y + h, p->height));
        while(h && y < p->height ) {
                for(i=0;i<w && x + i < p->width;i++) {
                       CELL(p,x + i,y) = style; 
                }
                
               h--;y++;
        }
}

void
pane_touchall(struct pane *p)
{
        //p->redraw_all = 1;
        //pane_touchcell(p,0,0);
        //pane_touchcell(p,p->width,p->height);
}

void 
pane_toucharea(struct pane *p, int x, int y, int w, int h)
{
        if(!w || !h) return;
        pane_touchcell(p, (x)/p->cwidth, (y)/p->cheight);
        pane_touchcell(p, (x + w)/p->cwidth, (y + h)/p->cheight);
}

void 
pane_touchcell(struct pane *p, int x, int y) 
{
        assert(x >= 0 && y >= 0);
        if(p->rr.x1 == -1 ) {
                p->rr.x1 = p->rr.x2 = x;
                p->rr.y1 = p->rr.y2 = y;
        } else {
                p->rr.x1 = MIN(p->rr.x1, x);
                p->rr.y1 = MIN(p->rr.y1, y);
                p->rr.x2 = MAX(p->rr.x2, x);
                p->rr.y2 = MAX(p->rr.y2, y);
        }
}

void
pane_showcursor(struct pane *p, bool show)
{
        if(p->show_cursor == show) return;
        p->show_cursor = show;
        pane_touchcell(p,p->cursor_x,p->cursor_y);
}

void 
pane_setcursor(struct pane *p, int x, int y)
{
        if(p->show_cursor) {
                pane_touchcell(p,p->cursor_x,p->cursor_y);
                pane_touchcell(p,x,y);
        }
        p->cursor_x = x;
        p->cursor_y = y;
        pane_sane(p);
}

static void
pane_sane(struct pane *p)
{
        assert(p->width >= 0); 
        assert(p->height >= 0); 
        assert(!p->width || p->cursor_x <= p->width);
        assert(!p->height || p->cursor_y < p->height);
        assert(!p->width || !p->height || (p->cells));
}

void
pane_tick(struct pane *p)
{
}

void 
pane_scroll(struct pane *p, int y)
{
        if(!y) return;
        if(abs(y) >= p->height) return;
        int sz = sizeof(struct cell)*p->width*(p->height - abs(y));
        if ( y > 0) 
                memmove(p->cells , p->cells + p->width*y,sz);
        else 
                memmove(p->cells + p->width*(-y), p->cells,sz);
        //p->redraw_all = 1;
        pane_touchcell(p,0,0);
        pane_touchcell(p,p->width,p->height);
}

void
pane_copy(struct pane *p, int sx, int sy, int dx, int dy, int w, int h)
{
        //fprintf(stderr, "pane_copy(%p, sx = %d, sy = %d, dx = %d, dy = %d, w = %d, h = %d\n", p, sx, sy, dx, dy, w, h);
        if (!h || !w) return;
        assert ( h > 0 && w > 0);
        pane_touchcell(p, dx, dy);
        pane_touchcell(p, MIN(dx + w, p->width), MIN(dy + h, p->height));
        if( sx == 0 && dx == 0 && w == p->width && abs(sy - dy) + h == p->height) {
                pane_scroll(p, sy - dy);
                return;
        }
        if (sy > dy)
                for (int i = 0; i < h; ++i) {
                        memmove(&CELL(p,dx,dy + i), &CELL(p, sx , sy + i), w*sizeof(struct cell));
                }
        else
                for (int i = h - 1; i >= 0; --i) {
                        memmove(&CELL(p,dx,dy + i), &CELL(p, sx , sy + i), w*sizeof(struct cell));
                }
}


void
pane_draw(struct pane *p) 
{
        int i,j;
        //if(p->redraw_all) {
        //        XSetForeground(disp, win.gc, BlackPixel(disp, DefaultScreen(disp)));
        //        XFillRectangle(disp, win.w, win.gc, 0, 0, win.width, win.height);
        //}
        //printf("pane_draw(%d,%d,%d,%d)\n", p->rr.x1, p->rr.y1, p->rr.x2 - p->rr.x1 + 1, p->rr.y2 - p->rr.y1 + 1);
        
        if (p->rr.x1 == -1)
                return;

        
        for(i = 0; i < p->width; i++)
                for(j = 0; j < p->height; j++) {
                        struct cell *c = &CELL(p,i,j);
                        wchar_t ch = c->ch;
                        //if(!p->redraw_all) {
                        //        if (!TOUCHED(p,i,j)) continue; 
                        //ll        if (ch == 0) ch = ' ';
                        //} //else { if (ch == 0) continue; }
                        //if (ch == 0 && TOUCHED(p,i,j)) ch = ' '; else continue;
                        if(i < p->rr.x1 || i > p->rr.x2) continue;
                        if(j < p->rr.y1 || j > p->rr.y2) continue;
                        if (ch == 0) ch = ' ';
                        XChar2b b;
                        b.byte2 = ch&0xff;
                        b.byte1 = (ch >> 8)& 0xFF;
                        XSetForeground(disp, win.gc, c->flag.inverse ? c->bgcolor : c->fgcolor);
                        XSetBackground(disp, win.gc, c->flag.inverse ? c->fgcolor : c->bgcolor);
                        XDrawImageString16(disp, win.w, win.gc, i*p->cwidth, j*p->cheight + xfs->ascent , &b, 1);
                }
        if(p->show_cursor) {
                XSetForeground(disp, win.gc, p->cursor_color);
                XDrawRectangle(disp, win.w, win.gc, p->cwidth * p->cursor_x, p->cheight * p->cursor_y, p->cwidth - 1, p->cheight - 1 );
        }
        //XCopyArea(disp, win.pm, win.w, win.gc, 
        //        0, 0, win.width, win.height, 0,0);    
        //memset(p->touched, 0, p->width*p->height);
        //p->redraw_all = 0;
        //p->any_touched = 0;
        p->rr.x1 = p->rr.x2 = -1;
        p->rr.y1 = p->rr.y2 = -1;
}
        


