#include "config.h"

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <pty.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <utmp.h>
#include <wchar.h>
#include <wctype.h>

#include <X11/Xlocale.h>
#include <X11/keysym.h>

#include "defaults.h"
#include "emulator.h"
#include "keysym2ucs.h"
#include "multigraph.h"
#include "oop.h"
#include "pane.h"
#include "panic.h"
#include "print_event.h"
#include "unicode.h"
#include "unix.h"
#include "x11.h"
#include "xatoms.h"
/* arch-tag: f5eee741-ae16-49fc-b63c-7c7da7ffed3f */


static void queue(unsigned what);

static bool is_visible;
static oop_source_sys *oss;
static oop_source *os;

static int num_pending = 0;
static int num_final = 0;

static struct pty *pty;
static struct pane *pane;
static struct emulate *emulate;

#define PEND_QCALLBACK 1
#define PEND_TIORESIZE 2
#define PEND_XFLUSH 4
#define PEND_XEXPOSE (PEND_XFLUSH | 8)

static unsigned pending;

static fifo_t out_fifo = FIFO_BLANK;
static fifo_t in_fifo = FIFO_BLANK;


typedef struct line line_t;

struct line {
        rb_t cs;
        rb_t colors;
}; 

static rb_t lines;

static void *
event_finalizer()
{
        num_final++;
        if(pending & PEND_TIORESIZE)
                ioctl(pty->fd, TIOCSWINSZ, &pty->winsize);
        if((pending & PEND_XEXPOSE) == PEND_XEXPOSE && is_visible) pane_draw(pane);
        if(pending & PEND_XFLUSH) XFlush(disp);
        pending = 0;
        if((num_final % 10) == 0)
                fprintf(stderr, "p/f = %i/%i  %lf\n", num_pending, num_final, (double)num_pending/(double)num_final);
        return OOP_CONTINUE;
}


static void
queue(unsigned what)
{
        struct timeval soon;
        assert(what);
        gettimeofday(&soon, NULL);
        //soon.tv_sec += 1;
        soon.tv_usec += 10;
        if(!pending)
                os->on_time(os, soon,  event_finalizer, NULL);
        pending |= what;
}

static void
resize_event(int w, int h)
{
        int fw, fh;
        int cwidth, cheight;
        if (w == win.width && h == win.height) return;
        win.width = w;
        win.height = h;
        get_charcell_size(&fw, &fh);
        if(DB_force80x25) {
                cheight = 25;
                cwidth = 80;
        } else {
                cheight = h / fh;
                cwidth = w / fw;
        }
        pty->winsize.ws_row = cheight;
        pty->winsize.ws_col = cwidth;
        pty->winsize.ws_xpixel = w;
        pty->winsize.ws_ypixel = h;

        pane_setup(pane, cwidth, cheight);
        emulate_set_size(emulate, cwidth, cheight);
        pane_clear(pane, 0, 0, cwidth, cheight, emulate->style);
        //XFreePixmap(disp, win.pm);
        //win.pm = XCreatePixmap(disp, win.w, w, h, DefaultDepth(disp, DefaultScreen(disp)));
        queue(PEND_TIORESIZE | PEND_XEXPOSE);
}


static void *
on_writeable(oop_source *source, int fd, oop_event event, void *_pty) 
{
        ssize_t n2;
        size_t n = FIFO_LEN(out_fifo);
        //fprintf(stderr, "on_writeable %u\n", n);
        if(n) {
                if((n2 = write(fd, FIFO_PTR(out_fifo), n)) < 0) {
                        if(errno == EINTR) 
                                return OOP_CONTINUE;
                        else
                                panicno("on_writeable, write");
                }
                if (n2 < n) {
                        FIFO_POP(out_fifo, n2);
                        //fprintf(stderr, "on_writeable wait %u\n", n - n2);
                        return OOP_CONTINUE;
                }
                FIFO_CLEAR(out_fifo);
        }
        os->cancel_fd(source, fd, OOP_WRITE);
        return OOP_CONTINUE;
}

static void
stuff_buf(struct pty *pty, char *buf, int len)
{
        if(FIFO_EMPTY(out_fifo)) 
                os->on_fd(os, pty->fd, OOP_WRITE, on_writeable, pty);
        fifo_append(&out_fifo, buf, len);
}


static void
send_wchar(struct pty * restrict pty, wchar_t ch)
{
        unsigned char buf[8];
        int n = utf8_wctomb(buf,ch,8);
        assert(n > 0);
        stuff_buf(pty, buf, n);
}

//TODO 
//fix overflow bugs.
static void
handle_keypress(XKeyEvent kev) 
{
        KeySym keysym;
        char *ptr = NULL;
        static char mg_buf[64];
        static char *mg_state = NULL;
        unsigned char buf[41];
        int n = XLookupString(&kev,buf,40,&keysym,NULL);
        if(n > 0) { buf[n] = '\0'; ptr = buf; }
        if(!mg_state && keysym == XK_2 && (kev.state & ControlMask)) {
                XSetWindowBorder(disp, win.w, DC_digraphBorderColor); 
                queue(PEND_XFLUSH);
                mg_state = mg_buf;
                memset(mg_buf, 0, sizeof(mg_buf));
                return;
        } 
        if(mg_state) {
                struct mgraph *mg;
                strcpy(mg_state, buf);
                mg_state += n; 
                mg = bsearch(mg_buf, mgraphs, NUM_MGRAPHS, sizeof(mgraph_t), compare_mgraph_exact); 
                if(mg) {
                        send_wchar(pty, mg->c);
                        mg_state = NULL;
                        XSetWindowBorder(disp, win.w, DC_borderColor); 
                        queue(PEND_XFLUSH);
                        return;
                }
                mg = bsearch(mg_buf, mgraphs, NUM_MGRAPHS, sizeof(mgraph_t), compare_mgraph_substr); 
                if(!mg) {
                        mg_state = NULL;
                        XSetWindowBorder(disp, win.w, DC_borderColor); 
                        queue(PEND_XFLUSH);
                }
                else
                        return;
        }
        if( keysym == XK_1 && (kev.state & ControlMask)) {
                char *s;
                asprintf(&s, "#num_pending %u\n#num_final %u\n#out_fifo %u %u\n#in_fifo %u %u\n", num_pending, num_final, FIFO_LEN(out_fifo), out_fifo.rb.size, FIFO_LEN(in_fifo), in_fifo.rb.size);
                stuff_buf(pty, s, strlen(s));
                free(s);
                return;
        }

#include "keys-gen.h"

        if(ptr) {
                for(;*ptr;ptr++) send_wchar(pty, *ptr);
        } else {
                unsigned cp = keysym2ucs(keysym);
                if (cp != -1) send_wchar(pty, cp);
        }
}

static void *
on_x_event(oop_source *source, int fd, oop_event event, void *null) {
        XEvent ev;

        while(XPending(disp)) {
                XNextEvent(disp, &ev);
                //print_event(disp, ev);
                switch(ev.type) {
                case ClientMessage:
                        if((Atom)ev.xclient.data.l[0]==XA_WM_DELETE_WINDOW)
                                return NULL;
                        print_event(disp, ev);
                        break;
                case Expose:
                        pane_toucharea(pane, ev.xexpose.x, ev.xexpose.y, ev.xexpose.width, ev.xexpose.height);
                        queue(PEND_XEXPOSE);  
                        break;
                case KeyPress:
                        handle_keypress(ev.xkey);
                        break;
                case MappingNotify:
                        XRefreshKeyboardMapping(&(ev.xmapping));
                        queue(PEND_XFLUSH);
                        break;
                case ConfigureNotify:
                        print_event(disp, ev);
                        resize_event(ev.xconfigure.width, ev.xconfigure.height);
                        break;
                case UnmapNotify:
                        is_visible = false;
                        break;
                case VisibilityNotify:
                        switch (ev.xvisibility.state) {
                        case VisibilityUnobscured:
                        case VisibilityPartiallyObscured:
                                is_visible = true;
                                break;
                        case VisibilityFullyObscured:
                                is_visible = false;
                                break;
                        default:
                                break;
                        }
                        break;
                default:
                        print_event(disp, ev);
                }
        }
        num_pending++;
        return OOP_CONTINUE;
}


#define INCOMING_BUFFER 4096

static void *
on_incoming_data(oop_source *source, int fd, oop_event event, void *null)
{
        unsigned char buf[INCOMING_BUFFER];
        wchar_t wc;
        int i,r;

        ssize_t ret = read(fd, buf, INCOMING_BUFFER);
        if(ret == -1) 
                panicno(pty->name);
        if(!ret)
                return NULL;
        //fprintf(stderr, "on_incoming_data %u %u\n", FIFO_LEN(in_fifo), ret);
        fifo_append(&in_fifo, buf, ret);
        do {
                char *p = FIFO_PTR(in_fifo);
                int len = FIFO_LEN(in_fifo);
                for(i = 0; i < len && !(p[i]&0x80); i++); 
                if (i) {
                        emulate_data_latin1(emulate, p, i, pane);
                        FIFO_POP(in_fifo, i);
                        len -= i;
                        p += i;
                }
                if(len) {
                        r = utf8_mbtowc(&wc,p,len);
                        if(!r) continue;
                        if(r < 0) {
                                r = -r;
                                fprintf(stderr, "invalid bytes:");
                                for (int x = 0; x < r; x++)
                                        fprintf(stderr, " %02x", (unsigned)((unsigned char *)FIFO_PTR(in_fifo))[x]);
                                fprintf(stderr, "\n");
                        } else 
                                emulate_data(emulate, &wc, 1, pane);
                        FIFO_POP(in_fifo, r);

                } else
                        r = 0;
        } while(!FIFO_EMPTY(in_fifo) && (i || r));
        num_pending++;
        queue(PEND_XEXPOSE); 
        return OOP_CONTINUE;
}




int
main(int argc,char *argv[])
{
        char *dargv[] = { "/bin/zsh", NULL };
        setlocale(LC_ALL, "");
	x11_init(argc, argv);	
        char *buf;

        pane = pane_create();
        emulate = emulate_create();


        oss = oop_sys_new();
        os = oop_sys_source(oss);


	XSelectInput(disp,win.w,KeyPressMask|ExposureMask|StructureNotifyMask|VisibilityChangeMask);


        unsetenv("LINES");
        unsetenv("COLUMNS");
        unsetenv("TERMCAP");
        putenv("TERM=xterm");
        asprintf(&buf, "WINDOWID=%lu", win.w);
        putenv(buf);


	pty = spawn_p(argc > 1 ? argv+1 : dargv);

        int fl = fcntl(pty->fd, F_GETFL);
        if(fl == -1) panicno("fcntl");
        fl |= O_NONBLOCK;
        fl = fcntl(pty->fd, F_SETFL, fl);
        if(fl == -1) panicno("fcntl");

        win.width = -1;
        resize_event(640,480);
        XSync(disp, False);
        on_x_event(0,0,0,0);

        XMapWindow(disp, win.w);

        os->on_fd(os, ConnectionNumber(disp), OOP_READ, on_x_event, NULL);
        os->on_fd(os, pty->fd, OOP_READ, on_incoming_data, NULL);


        oop_sys_run(oss);

	x11_close();
	return 0;
}

/*
void
output_char(wchar_t wc)
{
        switch(wc) {
        case '\n': c_newline(); return;
        case '\a': XBell(disp, 75); queue(PEND_XFLUSH); return;
        case '\t': cx = ((cx + 8) / 8) * 8; if (cx >= cwidth) c_newline(); pane_setcursor(pane, cx, cy);  return;
        case '\r': cx = 0;  pane_setcursor(pane, cx, cy); return;
        case '\b': if (cx > 0) { cx--; pane_setcursor(pane, cx, cy); } return;
        default: break;
        }
        if(!iswprint(wc)) {
                wchar_t buf[20];
                
                //char *c = buf;
                int len = swprintf(buf, 20, L"(%x)", (unsigned)wc);
                pane_settext(pane, cx, cy, buf, len);
                cx += len;
                if(cx >= cwidth) c_newline();
                pane_setcursor(pane, cx, cy);

        } else {
                if(cx == cwidth) c_newline();
                pane_settext(pane, cx++, cy, &wc, 1);
                pane_setcursor(pane, cx, cy);
        }
        queue(PEND_XEXPOSE);
}
static void 
c_newline()
{
        cx = 0;
        cy++;
        if (cy == cheight) {
                cy--;
                pane_scroll(pane, 1);
                pane_cleararea(pane, 0, cheight - 1, cwidth, 1, 0);
        }
        pane_setcursor(pane, cx, cy);
        queue(PEND_XEXPOSE);
}
#define UTF8_MB 2048


static void *
on_incoming_data(oop_source *source, int fd, oop_event event, void *null)
{
        static unsigned char buf[UTF8_MB];
        static unsigned i = 0;
        wchar_t wc;
        ssize_t ret;
        int r;
        assert(i >= 0 && i < UTF8_MB);

        ret = read(fd, buf + i, UTF8_MB - i);
        if(!ret)
                return NULL;
        if(ret == -1) 
                panicno(pty->name);
        
        assert(ret > 0 && ret <= UTF8_MB - i);
        i += ret;
        assert(i > 0 && i <= UTF8_MB);
        while(i > 0 && (r = utf8_mbtowc(&wc,buf,i)) != 0) {
                int x,ar = abs(r);
                assert (ar > 0 && abs(r) <= 6);
                assert (abs(r) <= i);
                if(r < 0) {
                        fprintf(stderr, "invalid bytes: ");
                        for (x = 0; x < abs(r); x++)
                                fprintf(stderr, " %x", (int)buf[x]);
                fprintf(stderr, "\n");
                }

                if (r > 0)
                        emulate_data(emulate, &wc, 1, pane);
                        //output_char(wc);
                r = abs(r);
                memmove(buf, buf + r, i - r);
                i-=r;
        }
        num_pending++;
        queue(PEND_XEXPOSE);  // new
        return OOP_CONTINUE;
}
*/
